@@ -47,11 +47,6 @@ try:
 except ModuleNotFoundError:
     py2neo = None
 
-try:
-    import neotime
-except ModuleNotFoundError:
-    neotime = None
-
 """TODO: Remove this after this is somewhat done"""
 import logging
 
@@ -666,7 +661,9 @@ class Graph(Widget):
             node_attributes : dictionary of node attributes
             """
             for k, v in node_attributes.items():
-                if neotime and isinstance(v, neotime.Date):
+                try:
+                    json.dumps(v)
+                except TypeError:
                     node_attributes[k] = str(v)
 
             return node_attributes
